// M5Kiss : Bibliothèque contenant des fonctions de conversion KISS, AX25 et TNC2.
// Ecrit par Frank LOBA (www.f4hdb.fr).
// Date création : 29/11/2022.
// Dernière mise à jour : 06/12/2022.


/**** DEBUT DE FICHIER ****/
#ifndef M5KISS_CPP
#define M5KISS_CPP


/**** INCLUSION DES FICHIERS EN-TETE ****/
#include "M5Kiss.h"


/**** DEFINITION DE LA FONCTION DE CONVERSION DU FORMAT TNC2 VERS LE FORMAT AX25 ****/
String  TNC2toAX25(String TrameTNC2)
{
  int     IndexTNC2 = 0;
  int     IndexRpt = 0, NbRpt = 0;
  String  Src = "";
  String  Dst = "";
  String  Rpt[AX25_MAX_RPT];
  String  Msg = "";
  char    Ssid;
  bool    IsRpt;
  String  TrameAX25 = "";

  // Lecture de la trame TNC2.

  // Détection d'une trame TNC2 valide.
  // Elle doit contenir le caractère TNC2_DST_CAR et le caractère TNC2_MSG_CAR.
  if(TrameTNC2.indexOf(TNC2_DST_CAR) == -1 || TrameTNC2.indexOf(TNC2_MSG_CAR) == -1)
  {
    TrameAX25 = "";
    return TrameAX25;
  }

  // Lecture de l'indicatif de la station source dans la trame TNC2.
  for(IndexTNC2 = 0; IndexTNC2 < TrameTNC2.indexOf(TNC2_DST_CAR); IndexTNC2 ++)
  Src += TrameTNC2.charAt(IndexTNC2);
  
  // Lecture de l'indicatif de la station de destination dans la trame TNC2.
  IndexTNC2 = TrameTNC2.indexOf(TNC2_DST_CAR) + 1;
  
  while(TrameTNC2.charAt(IndexTNC2) != TNC2_MSG_CAR && TrameTNC2.charAt(IndexTNC2) != TNC2_VIA_CAR)
  {
    Dst += TrameTNC2.charAt(IndexTNC2);
    IndexTNC2 ++;
  }

  // Lecture des indicatifs des répéteurs dans la trame TNC2.
  IndexRpt = 0;

  while(TrameTNC2.charAt(IndexTNC2) == TNC2_VIA_CAR && IndexRpt < AX25_MAX_RPT)
  {
    IndexTNC2 ++;
    Rpt[IndexRpt] = "";
    
    while(TrameTNC2.charAt(IndexTNC2) != TNC2_VIA_CAR && TrameTNC2.charAt(IndexTNC2) != TNC2_MSG_CAR)
    {
      Rpt[IndexRpt] += TrameTNC2.charAt(IndexTNC2);
      IndexTNC2 ++;
    }

    IndexRpt ++;
  }

  // Ecriture du nombre de stations répétitrices comptées dans la trame au format TNC2.
  NbRpt = IndexRpt;

  // Lecture du message.
  if(TrameTNC2.charAt(IndexTNC2) == TNC2_MSG_CAR)
  {
    Msg = TrameTNC2.substring(TrameTNC2.indexOf(TNC2_MSG_CAR) + 1);
  }
  else
  {
    TrameAX25 = "";
    return TrameAX25;
  }
  

  // Ecriture de la trame AX25.

  // Vérification de la présence d'un SSID dans l'indicatif de la station de destination au format TNC2.
  if(Dst.indexOf(TNC2_SSID_CAR) == -1)
  // On ajoute -0 s'il n'y a pas de SSID.
  Dst += "-0";

  // Vérification d'une taille correcte pour l'indicatif de la station de destination au format TNC2.
  if(Dst.indexOf(TNC2_SSID_CAR) > AX25_NAME_SIZE)
  {
    TrameAX25 = "";
    return TrameAX25;
  }
  
  // Ecriture de l'indicatif de la station de destination au format AX25.
  for(IndexTNC2 = 0; IndexTNC2 < AX25_NAME_SIZE; IndexTNC2 ++)
  {
    // Présence d'un caractère dans l'indicatif de la station de destination au format TNC2.
    if(IndexTNC2 < Dst.indexOf(TNC2_SSID_CAR))
    // On l'écrit dans la trame AX25 en le décalant les bits d'un rang vers la gauche pour mettre le bit 0 dit d'extention à 0.
    TrameAX25 += (char) (Dst.charAt(IndexTNC2) << 1);
    // Fin de l'indicatif de la station de destination au format TNC2.
    else
    TrameAX25 += (char) (AX25_SPACE << 1);
  }

  // Lecture du SSID de la station de destination au format TNC2.
  Ssid = (char) Dst.substring(Dst.indexOf(TNC2_SSID_CAR) + 1).toInt();
  // Vérification de la conformité du SSID.
  if(Ssid >= 16)
  {
    TrameAX25 = "";
    return TrameAX25;
  } 
  // Décalage des bits d'un rang vers la gauche pour mettre à 0 du bit 0 d'extention.
  Ssid <<= 1;
  // Masquage avec les bits réservés.
  Ssid |= AX25_SSID_RSD_MASK;
  // Ecriture du SSID de la station de destination au format AX25.
  TrameAX25 += (char) Ssid;


  // Vérification de la présence d'un SSID dans l'indicatif de la station source au format TNC2.
  if(Src.indexOf(TNC2_SSID_CAR) == -1)
  // On ajoute -0 s'il n'y a pas de SSID.
  Src += "-0";

  // Vérification d'une taille correcte pour l'indicatif de la station source au format TNC2.
  if(Src.indexOf(TNC2_SSID_CAR) > AX25_NAME_SIZE)
  {
    TrameAX25 = "";
    return TrameAX25;
  }
  
  // Ecriture de l'indicatif de la station source au format AX25.
  for(IndexTNC2 = 0; IndexTNC2 < AX25_NAME_SIZE; IndexTNC2 ++)
  {
    // Présence d'un caractère dans l'indicatif de la station source au format TNC2.
    if(IndexTNC2 < Src.indexOf(TNC2_SSID_CAR))
    // On l'écrit dans la trame AX25 en le décalant les bits d'un rang vers la gauche pour mettre le bit 0 dit d'extention à 0.
    TrameAX25 += (char) (Src.charAt(IndexTNC2) << 1);
    // Fin de l'indicatif de la station source au format TNC2.
    else
    TrameAX25 += (char) (AX25_SPACE << 1);
  }

  // Lecture du SSID de la station source au format TNC2.
  Ssid = (char) Src.substring(Src.indexOf(TNC2_SSID_CAR) + 1).toInt();
  // Vérification de la conformité du SSID.
  if(Ssid >= 16)
  {
    TrameAX25 = "";
    return TrameAX25;
  } 
  // Décalage des bits d'un rang vers la gauche pour mettre à 0 du bit 0 d'extention.
  Ssid <<= 1;
  // Masquage avec les bits réservés.
  Ssid |= AX25_SSID_RSD_MASK;
  // S'il n'y a pas de stations répétitrices on positionne le bit 0 d'exention à 1.
  if(NbRpt == 0) Ssid |= AX25_SSID_EXT_MASK;
  // Ecriture du SSID de la station source au format AX25.
  TrameAX25 += (char) Ssid;


  // Pour chaque station répétitrice.
  for(IndexRpt = 0; IndexRpt < NbRpt; IndexRpt ++)
  {
    // Vérification si la station doit répéter.
    IsRpt = (Rpt[IndexRpt].indexOf(TNC2_RPT_CAR) != -1);
    
    // Si la station répétitrice ne doit plus répéter, présence du caractère * à la fin de l'indicatif.
    if(IsRpt)
    // On supprime le caractère *.
    Rpt[IndexRpt] = Rpt[IndexRpt].substring(0, Rpt[IndexRpt].length() - 1);

    // Vérification de la présence d'un SSID dans l'indicatif de la station répétitrice au format TNC2.
    if(Rpt[IndexRpt].indexOf(TNC2_SSID_CAR) == -1)
    {
      // Dans tous les cas, on ajoute -0 s'il n'y a pas de SSID.
      Rpt[IndexRpt] += "-0";
    }
  
    // Vérification d'une taille correcte pour l'indicatif de la station répétitrice au format TNC2.
    if(Rpt[IndexRpt].indexOf(TNC2_SSID_CAR) > AX25_NAME_SIZE)
    {
      TrameAX25 = "";
      return TrameAX25;
    }
    
    // Ecriture de l'indicatif de la station répétitrice au format AX25.
    for(IndexTNC2 = 0; IndexTNC2 < AX25_NAME_SIZE; IndexTNC2 ++)
    {
      // Présence d'un caractère dans l'indicatif de la station répétitrice au format TNC2.
      if(IndexTNC2 < Rpt[IndexRpt].indexOf(TNC2_SSID_CAR))
      // On l'écrit dans la trame AX25 en le décalant les bits d'un rang vers la gauche pour mettre le bit 0 dit d'extention à 0.
      TrameAX25 += (char) (Rpt[IndexRpt].charAt(IndexTNC2) << 1);
      // Fin de l'indicatif de la station source au format TNC2.
      else
      TrameAX25 += (char) (AX25_SPACE << 1);
    }
  
    // Lecture du SSID de la station répétitrice au format TNC2.
    Ssid = (char) Rpt[IndexRpt].substring(Rpt[IndexRpt].indexOf(TNC2_SSID_CAR) + 1).toInt();
    // Vérification de la conformité du SSID.
    if(Ssid >= 16)
    {
      TrameAX25 = "";
      return TrameAX25;
    } 
    // Décalage des bits d'un rang vers la gauche pour mettre à 0 du bit 0 d'extention.
    Ssid <<= 1;
    // Masquage avec les bits réservés.
    Ssid |= AX25_SSID_RSD_MASK;
    // S'il n'y a plus de stations répétitrices on positionne le bit 0 d'exention à 1.
    if(IndexRpt == NbRpt - 1) Ssid |= AX25_SSID_EXT_MASK;
    // Si la station a déjà répété, on positionne le bit 7 correspondant à 1.
    if(IsRpt) Ssid |= AX25_SSID_RPT_MASK;
    // Ecriture du SSID de la station répétitrice au format AX25.
    TrameAX25 += (char) Ssid;   
  }


  // Ecriture de l'octet CONTROL FIELD UI dans la trame AX25.
  TrameAX25 += (char) AX25_CONTROL_FIELD_UI;
   
  // Ecriture de l'octet PROTOCOL ID dans la trame AX25.
  TrameAX25 += (char) AX25_PROTOCOL_ID;

  // Ecriture du message dans la trame AX25.
  TrameAX25 += Msg;

  // Fin de la fonction.
  return TrameAX25;
}


/**** DEFINITION DE LA FONCTION DE CONVERSION DU FORMAT AX25 VERS LE FORMAT TNC2 ****/
String  AX25toTNC2(String TrameAX25)
{
  int     IndexAX25 = 0, IndexLoop = 0;
  int     IndexRpt = 0, NbRpt = 0;
  String  Src = "";
  String  Dst = "";
  String  Rpt[AX25_MAX_RPT];
  String  Msg = "";
  char    Ssid;
  bool    IsRpt;
  String  TrameTNC2 = "";
  bool    FinAddr = false;

  // Lecture de la trame AX25.

  // Lecture de l'indicatif de la station de destination au format AX25.
  for(IndexAX25 = 0; IndexAX25 < AX25_NAME_SIZE; IndexAX25 ++)
  {
    // On écrit le caractère si ce n'est pas un espace.
    if((TrameAX25.charAt(IndexAX25) >> 1) != AX25_SPACE)
    Dst += (char) (TrameAX25.charAt(IndexAX25) >> 1);
  }

  // Lecture du SSID de la station de destination au format AX25.
  Ssid = (char) TrameAX25.charAt(IndexAX25);
  // Masquage du SSID.
  Ssid &= AX25_SSID_VAL_MASK;
  // Suppression du bit 0 d'extension en décalant d'un rang vers la droite les bites du SSID.
  Ssid >>= 1;
  // Conversion du SSID en caractères s'il est supérieur à 0.
  if(Ssid > 0)
  {
    // Ajout du caractère -.
    Dst += (char) TNC2_SSID_CAR;
    // Ajout du SSID.
    Dst += (int) Ssid;
  }
  

  // Incrémentation de l'index AX25.
  IndexAX25 ++;
  
  // Lecture de l'indicatif de la station source au format AX25.
  for(IndexLoop = IndexAX25; IndexLoop < (IndexAX25 + AX25_NAME_SIZE); IndexLoop ++)
  {
    // On écrit le caractère si ce n'est pas un espace.
    if((TrameAX25.charAt(IndexLoop) >> 1) != AX25_SPACE)
    Src += (char) (TrameAX25.charAt(IndexLoop) >> 1);
  }

  // Incrémentation de l'index AX25.
  IndexAX25 += AX25_NAME_SIZE;
  
  // Lecture du SSID de la station source au format AX25.
  Ssid = (char) TrameAX25.charAt(IndexAX25);
  // Détection de la fin de l'adressage grâce au bit 0 d'extension.
  FinAddr = (Ssid & AX25_SSID_EXT_MASK) == AX25_SSID_EXT_MASK;
  // Masquage du SSID.
  Ssid &= AX25_SSID_VAL_MASK;
  // Suppression du bit 0 d'extension en décalant d'un rang vers la droite les bites du SSID.
  Ssid >>= 1;
  // Conversion du SSID en caractères s'il est supérieur à 0.
  if(Ssid > 0)
  {
    // Ajout du caractère -.
    Src += (char) TNC2_SSID_CAR;
    // Ajout du SSID.
    Src += (int) Ssid;
  }


  // Initialisation de l'index des répéteurs.
  IndexRpt = 0;

  // Tant que ce n'est pas la fin de l'adressage.
  while(!FinAddr)
  {
    // Incrémentation de l'index AX25.
    IndexAX25 ++;

    // Initialisation de la chaîne de caractère du répéteur.
    Rpt[IndexRpt] = "";
  
    // Lecture de l'indicatif du répéteur au format AX25.
    for(IndexLoop = IndexAX25; IndexLoop < (IndexAX25 + AX25_NAME_SIZE); IndexLoop ++)
    {
      // On écrit le caractère si ce n'est pas un espace.
      if((TrameAX25.charAt(IndexLoop) >> 1) != AX25_SPACE)
      Rpt[IndexRpt] += (char) (TrameAX25.charAt(IndexLoop) >> 1);
    }
  
    // Incrémentation de l'index AX25.
    IndexAX25 += AX25_NAME_SIZE;
    
    // Lecture du SSID du répéteur au format AX25.
    Ssid = (char) TrameAX25.charAt(IndexAX25);
    // Détection de la fin de l'adressage grâce au bit 0 d'extension.
    FinAddr = (Ssid & AX25_SSID_EXT_MASK) == AX25_SSID_EXT_MASK;
    // Détection si la station a déjà répété.
    IsRpt = (Ssid & AX25_SSID_RPT_MASK) == AX25_SSID_RPT_MASK;
    // Masquage du SSID.
    Ssid &= AX25_SSID_VAL_MASK;
    // Suppression du bit 0 d'extension en décalant d'un rang vers la droite les bites du SSID.
    Ssid >>= 1;
    // Conversion du SSID en caractères s'il est supérieur à 0.
    if(Ssid > 0)
    {
      // Ajout du caractère -.
      Rpt[IndexRpt] += (char) TNC2_SSID_CAR;
      // Ajout du SSID.
      Rpt[IndexRpt] += (int) Ssid;
    }
    
    // Ajout du caractère * si la station a déjà répété.
    if(IsRpt)
    Rpt[IndexRpt] += (char) TNC2_RPT_CAR;

    // Incrémentation de l'index du répéteur.
    IndexRpt ++;
  }


  // Affectation du nombre de répéteurs comptés.
  NbRpt = IndexRpt;

  // Incrémentation de l'index AX25.
  IndexAX25 ++;

  // Vérification de la présence de l'octet CONTROL FIELD UI dans la trame AX25.
  if(TrameAX25.charAt(IndexAX25) != (char) AX25_CONTROL_FIELD_UI)
  {
    TrameTNC2 = "";
    return TrameAX25;
  } 

  // Incrémentation de l'index AX25.
  IndexAX25 ++;
  
  // Vérification de la présence de l'octet PROTOCOL ID dans la trame AX25.
  if(TrameAX25.charAt(IndexAX25) != (char) AX25_PROTOCOL_ID)
  {
    TrameTNC2 = "";
    return TrameAX25;
  }

  // Incrémentation de l'index AX25.
  IndexAX25 ++;
  
  // Lecture du message dans la trame AX25.
  Msg = TrameAX25.substring(IndexAX25);


  // Ecriture de la trame TNC2.

  // Ecriture de la station source dans la trame TNC2.
  TrameTNC2 = Src;

  // Ecriture du caractère > dans la trame TNC2.
  TrameTNC2 += (char) TNC2_DST_CAR;

  // Ecriture de la station de destination dans la trame TNC2.
  TrameTNC2 += Dst;

  // Ecriture des répéteurs dans la trame TNC2.
  for(IndexRpt = 0; IndexRpt < NbRpt; IndexRpt ++)
  {
    // Ecriture du caractère , dans la trame TNC2.
    TrameTNC2 += (char) TNC2_VIA_CAR;

    // Ecriture du répéteur dans la trame TNC2;
    TrameTNC2 += Rpt[IndexRpt];
  }

  // Ecriture du caractère : dans la trame TNC2.
  TrameTNC2 += (char) TNC2_MSG_CAR;

  // Ecriture du message dans la trame TNC2.
  TrameTNC2 += Msg;
  
  // Fin de la fonction.
  return TrameTNC2;
}
  

/**** DEFINITION DE LA FONCTION D'ENCODAGE AVEC LE PROTOCOLE KISS ****/
String  EncoderKISS(String TrameX)
{
  String TrameKISS;
  
  // Ecriture de la commande au format KISS FEND puis DATA.
  TrameKISS += (char) KISS_FEND;
  TrameKISS += (char) KISS_DATA;

    // Conversion.
   for(int IndexX = 0; IndexX < TrameX.length(); IndexX ++)
    {
      // Détection d'un caractère FEND au format X.
      if(TrameX.charAt(IndexX) == KISS_FEND)
      {
          // Ecrite du caractère au format KISS FESC puis TFEND.
          TrameKISS += (char) KISS_FESC;
          TrameKISS += (char) KISS_TFEND;
      }
      
      // Détection d'un caractère FESC au format X.
      else if(TrameX.charAt(IndexX) == KISS_FESC)
      {
          // Ecrite du caractère au format KISS FESC puis TFESC.
          TrameKISS += (char) KISS_FESC;
          TrameKISS += (char) KISS_TFESC;
      }
      // Le caractère n'est pas spécial.
      else
      {
        // Ecriture du caractère sans le modifier.
        TrameKISS += TrameX.charAt(IndexX);   
      }
    }

  // Ecriture de la commande au format KISS FEND.
  TrameKISS += (char) KISS_FEND;

  // Fin de la fonction.
  return TrameKISS;
}


/**** DEFINITION DE LA FONCTION D'ENCODAGE AVEC LE PROTOCOLE KISS ****/
String  DecoderKISS(String TrameKISS)
{
  String TrameX;

  // Vérification de la présence de l'octet KISS FEND, KISS_DATA au début et KISS FEND à la fin de la trame KISS.
  if(TrameKISS.charAt(0) != (char) KISS_FEND || TrameKISS.charAt(1) != (char) KISS_DATA || TrameKISS.charAt(TrameKISS.length() - 1) != (char) KISS_FEND)
  {
    TrameX = "";
    return TrameX;
  }
  
   // Conversion.
   for(int IndexKISS = 2; IndexKISS < TrameKISS.length() - 1; IndexKISS ++)
    {
      // Détection d'un caractère FESC puis TFEND dans la trame KISS.
      if(TrameKISS.charAt(IndexKISS) == KISS_FESC && TrameKISS.charAt(IndexKISS + 1) == KISS_TFEND)
      {
          // Ecrite du caractère FEND dans la trame X.
          TrameX += (char) KISS_FEND;
          
          // Incrémentation de l'index KISS.
          IndexKISS ++;
      }
      // Détection d'un caractère FESC puis TFESC dans la trame KISS.
      else if(TrameKISS.charAt(IndexKISS) == KISS_FESC && TrameKISS.charAt(IndexKISS + 1) == KISS_TFESC)
      {
          // Ecrite du caractère FEND dans la trame X.
          TrameX += (char) KISS_FESC;
          
          // Incrémentation de l'index KISS.
          IndexKISS ++;
      }
      // Le caractère n'est pas spécial.
      else
      {
        // Ecriture du caractère sans le modifier.
        TrameX += TrameKISS.charAt(IndexKISS);   
      }
    }

  // Fin de la fonction.
  return TrameX;
}


/**** FIN DE FICHIER ****/
#endif
